"use strict";

var _require = require('./utils'),
    getConstCache = _require.getConstCache,
    var2Expression = _require.var2Expression,
    expr2str = _require.expr2str,
    getPackage = _require.getPackage;

module.exports = function (_ref) {
  var t = _ref.types;

  function IdentifierVisitor(path, _ref2) {
    var opts = _ref2.opts,
        cache = _ref2.cache;
    var parent = path.parent;
    if (!parent) return;
    if (['FunctionDeclaration', 'ClassMethod', 'ObjectMethod'].includes(parent.type)) return;
    if (parent.type === 'ObjectProperty' && parent.key === path.node) return;
    if (parent.type === 'VariableDeclarator' && parent.id === path.node) return;
    var identifier = expr2str(path.node);

    if (parent.type === 'MemberExpression') {
      var parentPath = path.parentPath;

      while (parentPath && parentPath.node.type === 'MemberExpression') {
        if (path.node === parentPath.node.property) identifier = "".concat(expr2str(parentPath.node.object), ".").concat(identifier);
        path = parentPath;
        parentPath = parentPath.parentPath;
      }
    } else if (parent.type === 'CallExpression' && expr2str(parent.callee) === identifier) {
      if (identifier !== '__packageversion' || !parent.arguments.length || !t.isStringLiteral(parent.arguments[0])) return;
      path = path.parentPath;
      var packageName = expr2str(path.node.arguments[0]);

      var packagePath = packageName && require.resolve(packageName);

      var pkg = getPackage(packagePath);
      path.replaceWith(t.stringLiteral(pkg ? pkg.version : ''));
      return;
    }

    var defines = opts && opts.defines || {};

    if (identifier === '__filename') {
      path.replaceWith(t.stringLiteral(cache.filename));
    } else if (identifier === '__dirname') {
      path.replaceWith(t.stringLiteral(cache.dirname));
    } else if (identifier === '__now') {
      path.replaceWith(t.stringLiteral(cache.now));
    } else if (identifier === '__timestamp') {
      path.replaceWith(t.numericLiteral(Date.now()));
    } else if (defines[identifier] !== undefined) {
      path.replaceWith(var2Expression(defines[identifier]));
    } else if (cache.pkg) {
      if (identifier === '__packagename') {
        path.replaceWith(t.stringLiteral(cache.pkg.name));
      } else if (identifier === '__packageversion') {
        path.replaceWith(t.stringLiteral(cache.pkg.version));
      }
    }
  }

  return {
    visitor: {
      Program: {
        enter: function enter(path, state) {
          var filename = state.file.opts.filename,
              opts = state.opts;
          var cache = getConstCache(filename);
          path.traverse({
            Identifier: IdentifierVisitor
          }, {
            cache: cache,
            opts: opts
          });
        }
      }
    }
  };
};